//
//
//  Copyright 2013, Richard Gorton,     All Rights Reserved.
//
/// \file
/// Implementation of LWP context and ring buffer manipulation
///
/// \author Richard Gorton
///
/// \note:  For this particular implementation, have the constructor
///         allocate both the LWPCB and the Ring Buffer.  Another
///         implementation might choose a different policy.
///
///     The ring buffer can be located anywhere, but the spec recommends
///     the start of a page boundary.  We will presume this to be 4K
///     The ring buffer itself will be shared memory created with mmap
///

#include <malloc.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/mman.h>

#include "lwpcore.h"
#include "context.h"


static const size_t s_pageSize = 4096;

// Initialization function
// true if the initialization was successful
// We really only want to run this once per object
LWP_STATUS LWPContext::Initialize()
{
    LWP_STATUS  status = LWPStatus_Success;
    static bool initDone = false;
    bool    cleanupLWPCB = false;
    bool    cleanupRingBuffer = false;

    if (initDone)
    {
        goto done;
    }

    // fail if LWP is not available in hardware and supported by the OS
    if (!(LWP_FLAG_AVAILABLE & LWPHardwareFeatures()))
    {
        status = LWPStatus_LWPNoHardware;
        goto done;
    }

    if (!(LWP_FLAG_AVAILABLE & LWPOSFeatures()))
    {
        status = LWPStatus_LWPNotEnabledbyOS;
        goto done;
    }

    m_LWPCB = reinterpret_cast<LWPCB*>(memalign(s_pageSize, LWPCBSizeInBytes()));
    if (NULL == m_LWPCB)
    {
        // unrecoverable.
        status = LWPStatus_AllocationFailed;
        goto done;
    }

    m_EventConfigBase = reinterpret_cast<LWPCB_Event*>(m_LWPCB) + LWPEventOffset();

    // LWPCB Flags - set up separately

    // Ring buffer
    // Sizing first
    m_ringBufferBytes = m_ringBufferEvents * sizeof(LWP_GenericEvent);
    status = LWPSetRingBufferByteSize(m_LWPCB, m_ringBufferBytes);
    if (LWPStatus_Success != status)
    {
        // 28 bits max.  Asked for too much.  Bail out
        cleanupLWPCB = true;
        goto done;
    }

    status = CreateRingBuffer();
    if (LWPStatus_Success != status)
    {
        // unrecoverable.  Cleanup first
        cleanupLWPCB = true;
        goto done;
    }

    status = LWPSetRingBufferBaseAddress(m_LWPCB, m_ringBuffer);
    if (LWPStatus_Success != status)
    {
        // unrecoverable
        cleanupLWPCB = true;
        cleanupRingBuffer = true;
        goto done;
    }

    // Threshold - set up separately

    // Filters - set up separately

    // BaseIP - set up separately

    // LimitIP - set up separately

    // BufferTailOffset - initialized by constructor

    initDone = true;
done: ;
    if (cleanupRingBuffer)
    {
        ReleaseRingBuffer();
    }
    if (cleanupLWPCB)
    {
        free(m_LWPCB);
        m_LWPCB = NULL;
    }
    return status;
}

// create the ring buffer.  In this case, it is an mmap
// only the producer thread should do this
LWP_STATUS LWPContext::CreateRingBuffer()
{
    LWP_STATUS  status = LWPStatus_Success;

    // This could be a file backed item if the collector/harvester thread
    // needed to be in a separate process.
    m_ringBuffer = reinterpret_cast<LWP_GenericEvent*>(mmap(NULL,
                                                            m_ringBufferBytes,
                                                            PROT_READ|PROT_WRITE,
                                                            MAP_ANONYMOUS,
                                                            0,
                                                            0));
    if (MAP_FAILED == m_ringBuffer)
    {
        status = LWPStatus_AllocationFailed;
    }

    return status;
}

// Release the ring buffer.  In this case, it is an munmap
LWP_STATUS LWPContext::ReleaseRingBuffer()
{
    LWP_STATUS  status = LWPStatus_Success;
    int rc;

    if (NULL == m_ringBuffer)
    {
        status = LWPStatus_BadRingBufferAction;
        goto done;
    }

    rc = munmap(reinterpret_cast<void*>(m_ringBuffer), m_ringBufferBytes);
    if (0 != rc)
    {
        status = LWPStatus_BadRingBufferAction;
        goto done;
    }
    m_ringBuffer = NULL;

done: ;
    return status;
}
