//
//
//  Copyright 2013, Richard Gorton,     All Rights Reserved.
//
/// \file
/// A simple little utility to check to see which features of LWP are available both
/// in hardware, and fully supported by the OS.
/// \author Richard Gorton
///
/// \note   This implementation is 64-bit only.
///         Linux: not suppored by default in the kernel
///         Windows8: fully supported (including threshold mode) in the kernel
///         LWP REQUIRES the use of the XSAVE/XRSTOR instructions for context switching
///         See section 13.4.7 of the specification for details.

#if defined(_WIN32)
#error  Needs development work to support windows (requires Win8)
#endif

#include <stdint.h>
#include <iostream>
#include <string>

#include "lwpcore.h"

using namespace std;


/// Check for a particular feature flag, convert into yes/no string for
/// both hardware and OS support
/// \param[out]     string containing the yes/no values
/// \param[in]      bitfield mask
static void ReportFeature(std::string& outStr, uint32_t mask)
{
    outStr = "\t";
    if ((mask & LWPHardwareFeatures()) == mask)
    {
        outStr += "Yes";
    }
    else
    {
        outStr += "No";
    }
    outStr += "\t\t";
    if ((mask & LWPOSFeatures()) == mask)
    {
        outStr += "Yes";
    }
    else
    {
        outStr += "No";
    }
}

int main()
{

    LWPInit();

    // Now check the values in g_LWPCharacteristics
    std::string primaryFeatures;

    if ((LWP_FLAG_AVAILABLE & LWPHardwareFeatures()) == LWP_FLAG_AVAILABLE)
    {
        cout << "LWP Version: " << LWPVersion() << endl;
    }
    else
    {
        cout << "LWP not supported by the hardware" << endl;
    }

    cout << "Feature:\t\t\tHardware\tSoftware\n"
         << "\t\t\t\tSupport\t\tSupport"
         << endl;

    ReportFeature(primaryFeatures, LWP_FLAG_AVAILABLE);
    cout << "General LWP features:\t" << primaryFeatures << endl;

    cout << "Optional LWP features:" << endl;

    ReportFeature(primaryFeatures, LWP_FLAG_VALUESAMPLE);
    cout << "\tLWPVAL:\t\t" << primaryFeatures << endl;

    ReportFeature(primaryFeatures, LWP_FLAG_INSNRETIRED);
    cout << "\tInstructions Retired:" << primaryFeatures << endl;

    ReportFeature(primaryFeatures, LWP_FLAG_BRANCHRETIRED);
    cout << "\tBranches Retired:" << primaryFeatures << endl;

    ReportFeature(primaryFeatures, LWP_FLAG_DCACHE_MISS);
    cout << "\tDCached Misses:\t" << primaryFeatures << endl;

    ReportFeature(primaryFeatures, LWP_FLAG_CPU_NOT_HALTED);
    cout << "\tCPU Clocks not halted:" << primaryFeatures << endl;

    ReportFeature(primaryFeatures, LWP_FLAG_CONTINUOUS);
    cout << "\tContinuous mode:" << primaryFeatures << endl;

    ReportFeature(primaryFeatures, LWP_FLAG_SAVE_PTSC);
    cout << "\tSave PTSC:\t" << primaryFeatures << endl;

    ReportFeature(primaryFeatures, LWP_FLAG_THRESHOLD_INT);
    cout << "\tThreshold Interrupts:" << primaryFeatures << endl;

    std::string HWSupport("No");
    std::string OSSupport("No");

    if (((LWP_FLAG_DCACHE_MISS & LWPHardwareFeatures()) == LWP_FLAG_DCACHE_MISS)
        && LWPCacheLevelFiltering())
    {
        HWSupport = "Yes";
        if ((LWP_FLAG_DCACHE_MISS & LWPOSFeatures()) == LWP_FLAG_DCACHE_MISS
            && LWPCacheLevelFiltering())
        {
            OSSupport = "Yes";
        }
    }
    cout << "Cache Level filters:\t\t"
         << HWSupport
         << "\t\t"
         << OSSupport
         << endl;

    HWSupport = "No";
    OSSupport = "No";
    if (((LWP_FLAG_BRANCHRETIRED & LWPHardwareFeatures()) == LWP_FLAG_BRANCHRETIRED)
        && LWPBranchPredictionFiltering())
    {
        HWSupport = "Yes";
        if (((LWP_FLAG_BRANCHRETIRED & LWPOSFeatures()) == LWP_FLAG_BRANCHRETIRED)
            && LWPBranchPredictionFiltering())
        {
            OSSupport = "Yes";
        }
    }

    cout << "Branch prediction filters:" << endl;
    cout << "\tMispredicted:\t\t"
         << HWSupport
         << "\t\t"
         << OSSupport
         << "\n\tPredicted:\t\t"
         << HWSupport
         << "\t\t"
         << OSSupport
         << endl;

    ReportFeature(primaryFeatures, LWP_FLAG_BRANCHRETIRED);
    cout << "Branch type filters:" << endl;
    cout << "\tAbsolute:\t" << primaryFeatures << endl;
    cout << "\tConditional:\t" << primaryFeatures << endl;
    cout << "\tRelative:\t" << primaryFeatures << endl;

    HWSupport = "No";
    OSSupport = "No";
    if (LWPIPFiltering())
    {
        HWSupport = "Yes";
        OSSupport = "Yes";
    }
    cout << "IP filter:\t\t\t"
         << HWSupport
         << "\t\t"
         << OSSupport
         << "\nIP Invert filter:\t\t"
         << HWSupport
         << "\t\t"
         << OSSupport
         << endl;

    cout << "\nMinimum Ring Buffer size (units): "
         << LWPMinBufferUnits()
         << " == "
         << LWPMinBufferSizeInBytes()
         << " bytes"
         << endl;
    return 0;
}
